"""Implement a PCA to reduce the dimensionality of on image."""
import os

import imageio.v2 as imageio
import matplotlib.pyplot as plt
import numpy as np
from skimage.io import imsave
from skimage.metrics import structural_similarity
from tqdm import tqdm

np.random.seed(0)


def pca_transform(
    img_rows: np.ndarray,
) -> tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    """Perform a PCA transformation on the input matrix.

    Args:
        img_rows (np.ndarray): Rows of flattened image data (size n x m).

    Returns:
        tuple: Instances of the Principal Component Analysis:
        np.ndarray (m x 1): The eigenvalues sorted in descending order.
        np.ndarray (m x m): The eigenvectors sorted according to the order of the eigenvalues.
        np.ndarray (n x m): The zero-meaned data matrix.
        np.ndarray (1 x m): The vector of the feature means of the input matrix.
    """
    # 4. compute mean for each feature
    mean_v = np.mean(img_rows, axis=0)
    # 5. center input data by subtracting mean
    data_centered = img_rows - mean_v
    # 6. calculate covariance matrix
    covariance_matrix = np.cov(data_centered, rowvar=False)
    # 7. perform eigendecomposition of covariance matrix
    eigenval, eigenvec = np.linalg.eigh(covariance_matrix)
    # 8. sort eigenvalues in descending order and eigenvectors by descending eigenvalues
    sorted_indices = np.argsort(eigenval)[::-1]
    eigenvec = eigenvec[:, sorted_indices]
    eigenval = eigenval[sorted_indices]
    # 9. return sorted eigenvalues, eigenvectors, centered data and mean vector
    return eigenval, eigenvec, data_centered, mean_v


def pca_inverse_transform(
    data_centered: np.ndarray, eigenvec: np.ndarray, mean_v: np.ndarray, n_comp: int
) -> np.ndarray:
    """Perform PCA inverse transformation to reconstruct compressed data.

    Args:
        data_centered (np.ndarray): Data with zero mean.
        eigenvec (np.ndarray): The eigenvectors sorted according to the order of the PCA eigenvalues.
        mean_v (np.ndarray): The vector of the feature means.
        n_comp (int): The number of components for the reconstruction.

    Returns:
        np.ndarray: The reconstructed data.
    """
    # 10. select top 'n_comp' eigenvectors
    selected_eigenvectors = eigenvec[:, :n_comp]

    # 11. project centered data onto space defined by selected eigenvectors
    compressed_data = np.dot(data_centered, selected_eigenvectors)

    # 12. reconstruct reduced data
    reconstructed_data = np.dot(compressed_data, selected_eigenvectors.T) + mean_v
    # 13. return reconstructed data
    return reconstructed_data


def expl_var(eigenval: np.ndarray, n_comp: int) -> float:
    """Calculate the cumulative explained variance ratio for a given number of components.

    Args:
        eigenval (np.ndarray): The eigenvalues sorted in descending order.
        n_comp (int): The number of components.

    Returns:
        float: Cumulative explained variance ratio.
    """
    # 14. compute total variance
    total_variance = np.sum(eigenval)
    # 15. compute cumulative explained variance
    cumulative_explained_var = np.sum(eigenval[:n_comp])
    # 16. determine and return cumulative explained variance ratio
    cum_explained_variance_ratio = cumulative_explained_var / total_variance
    print(
        f"Explained Variance Ratio (n_comp={n_comp}): {cum_explained_variance_ratio:.2%}"
    )
    return cum_explained_variance_ratio


if __name__ == "__main__":
    # 1. create empty output dir
    os.makedirs("./output", exist_ok=True)

    # 2. load image, plot it and save as .png file
    image = imageio.imread("./data/images/statue.jpg")
    plt.imshow(image)
    plt.show()
    output_filename = "./output/original.png"
    imsave(output_filename, image.astype(np.uint8))
    plt.close()

    # 3. reshape image in 2D-array of shape (num_rows, num_columns * num_channels)
    image_rows = np.reshape(image, (image.shape[0], -1))

    # 17. compute PCA for all possible values of k
    ssims = []
    cumulative_explained_vars = []
    max_num = np.minimum(image.shape[0], image.shape[1])
    # 17.1. perform PCA using 'pca_transform' function
    eigenvalues, eigenvectors, centered_data, mean_vector = pca_transform(
        image_rows
    )
    # 17. iterate through range of n_components values (possibly incrementing by 10)
    for n_components in tqdm(range(0, max_num, 10)):
        # 17.2. reconstruct image from  lower-dimensional representation
        # using current number of components using 'pca_inverse_transform'
        recovered_image = pca_inverse_transform(
            centered_data, eigenvectors, mean_vector, n_components
        )
        # 17.3. reshape recovered image to its original shape
        # and save it in output folder
        recovered_image = np.reshape(recovered_image, image.shape)
        recovered_image = recovered_image.astype(np.uint8)
        plt.title(f"n_components = {n_components}")
        plt.imshow(recovered_image)
        output_filename = f"./output/pca_{n_components}.png"
        imsave(output_filename, recovered_image)
        plt.close()

        # 17.4. compute and store cumulative explained variance ratio
        # for current number of components using 'expl_var' function
        cumulative_explained_vars.append(expl_var(eigenvalues, n_components))

        # 17.5 compute and store SSIM
        ssims.append(
            structural_similarity(
                image, recovered_image, multichannel=True, channel_axis=2
            )
        )

    # 18. plot cumulative explained variance ratios...
    # create list of x-axis tick locations (every 10th component)
    x = np.arange(len(cumulative_explained_vars))
    x = x * 10

    fig, ax1 = plt.subplots()
    ax1.plot(x, cumulative_explained_vars)
    ax1.set_xlabel("number of components")
    ax1.set_ylabel("explained variance", color="C0")
    ax1.tick_params(axis="y", labelcolor="C0")

    # 19. ...and SSIM
    ax2 = ax1.twinx()
    ax2.plot(x, ssims, c="C1")
    ax2.set_ylabel("SSIM", color="C1")
    ax2.tick_params(axis="y", labelcolor="C1")

    plt.savefig("./output/explained_variance_ssim.png")
    plt.show()
