"""Use PCA to improve soft-margin SVM for face recognition."""
from time import time

import matplotlib.pyplot as plt
import numpy as np
from sklearn import svm
from sklearn.datasets import fetch_lfw_people
from sklearn.decomposition import PCA
from sklearn.metrics import accuracy_score
from sklearn.model_selection import GridSearchCV, KFold, train_test_split
from sklearn.preprocessing import StandardScaler

from util_pca import plot_image_matrix, plot_roc


# import or paste here your function cv_svm
def cv_svm(xtrain: np.ndarray, ytrain: np.ndarray) -> GridSearchCV:
    """Train and cross-validate a soft-margin SVM classifier with the grid search.

    Define an SVM classifier. Use the grid search and a 5-fold cross-validation
    to find the best value for the hyperparameters 'C' and 'kernel'.

    Args:
        xtrain (np.ndarray): The training data.
        ytrain (np.ndarray): The training labels.

    Returns:
        GridSearchCV: The trained model that was cross-validated with the grid search.
    """
    # define dictionary with parameter grids
    param_grid = {
        "C": [10**i for i in range(-2, 4, 1)],
        # 'gamma': [0.1, 1, 'scale', 'auto'],  # Kernel coefficient
        "kernel": ["rbf", "linear", "poly"],
    }
    # initialize svm classifier and perform grid search
    clf = GridSearchCV(svm.SVC(probability=True), param_grid)
    clf = clf.fit(xtrain, ytrain)
    # print parameters found with cross-validation
    print("Best estimator found with parameters:", clf.best_params_)
    return clf


def explained_var(xtrain: np.ndarray) -> np.ndarray:
    """Compute and plot the cumulative explained variance ratio for PCA.

    Calculate and plot the cumulative explained variance ratio for
    PCA applied to the training data.


    Args:
        xtrain (np.ndarray): The training data.

    Returns:
        np.ndarray: An array containing the cumulative explained variance ratios.
    """
    # 4. initialize PCA and fit on train data
    pca = PCA(whiten=True, random_state=42).fit(xtrain)

    # 5. plot cumulative explained variance ratios
    # of each principal component against the number of components
    cumulative_explained_var = np.cumsum(pca.explained_variance_ratio_)
    fig, ax1 = plt.subplots()
    ax1.plot(cumulative_explained_var)
    ax1.set_xlabel("number of components")
    ax1.set_ylabel("explained variance", color="C0")
    ax1.tick_params(axis="y", labelcolor="C0")
    plt.show()
    # 6. return array of cumulative explained variance ratios
    return cumulative_explained_var


def pca_train(
    xtrain: np.ndarray, ytrain: np.ndarray, n_comp: int, train_fun: GridSearchCV
) -> tuple[PCA, GridSearchCV]:
    """Train a model with a given train function on the PCA-transformed data.

    Extract the top 'n_comp' principal components from the training data using PCA,
    and then trains a model on these components as features.

    Args:
        xtrain (np.ndarray): The training data.
        ytrain (np.ndarray): The training labels.
        n_comp (int): The number of PCA components.
        train_fun: The estimator function used to train the model.

    Returns:
        tuple: A tuple containing the PCA decomposition
               with 'n_comp' components and the model trained with the 'train_fun' function.
    """
    # 8. initialize PCA and fit on train data
    pca = PCA(n_components=n_comp, whiten=True, random_state=42).fit(xtrain)

    # 9. transform input data using PCA transform
    x_train_pca = pca.transform(xtrain)
    # 10. train model on transformed PCA features
    model_pca = train_fun(x_train_pca, ytrain)
    # 11. return PCA decomposition object and trained model
    return pca, model_pca


# (optional)
def gs_pca(x: np.ndarray, y: np.ndarray, comps: np.ndarray) -> int:
    """Perform grid search to find the best number of PCA components.

     Conduct a grid search to identify the optimal number of PCA components
     to use for feature dimensionality reduction before the training.

    Args:
        x (np.ndarray): The input data.
        y (np.ndarray): The target labels.
        comps (np.ndarray): Array of numbers of PCA components to consider.

    Returns:
        int: The best number of PCA components.
    """
    # 20. define outer 5-fold cross-validation strategy
    outer_cv = KFold(n_splits=5, shuffle=True, random_state=42)  # 2

    # 21. initialize variables to track best score and components
    best_score = 0
    best_components = None
    # 22. loop over specified number of PCA components to consider
    for n_comp in comps:
        # 22.1. create outer cross-validation loop iterating over splits
        acc = []
        for train_index, test_index in outer_cv.split(x):
            # 22.1.1 take indices and generate current train and test set out of given data
            x_train_out, x_test_out = x[train_index], x[test_index]
            y_train_out, y_test_out = y[train_index], y[test_index]

            # 22.1.2. fit 'StandardScaler' on train data
            scaler = StandardScaler()
            scaler.fit(x_train_out)
            # 22.1.2. scale both train and test data
            x_train_out = scaler.transform(x_train_out)
            x_test_out = scaler.transform(x_test_out)
            # 22.1.3. create `PCA` instance with same parameters as before
            # and transform train data
            pca = PCA(n_components=n_comp, whiten=True, random_state=42).fit(
                x_train_out
            )
            x_train_pca = pca.transform(x_train_out)

            # 22.1.4. here either call 'cv_cvm' function
            # or initialize SVM classifier directly with parameters
            clf = svm.SVC(C=10, kernel="rbf", probability=True)
            clf = clf.fit(x_train_pca, y_train_out)
            # 22.1.5. predict labels on test data and compute accuracy score for each fold
            x_test_pca = pca.transform(x_test_out)
            y_pred = clf.predict(x_test_pca)
            acc.append((accuracy_score(y_pred, y_test_out) * 100).round(1))

        # 22.2. calculate mean accuracy score  across all folds
        # for current number of components
        acc_mean = sum(acc) / len(acc)
        print("Mean: ", acc_mean)
        # 22.3. check if this is best number of components based
        # on mean accuracy and update if needed
        if acc_mean > best_score:
            best_score = acc_mean
            best_components = n_comp
    # 23. return best number of PCA components found
    return best_components


if __name__ == "__main__":
    # 1. load dataset 'Labeled Faces in the Wild' and get data,
    # take only classes with at least 70 images; downsize images for speed up
    lfw_people = fetch_lfw_people(min_faces_per_person=70, resize=0.4)

    n_samples, h, w = lfw_people.images.shape
    x = lfw_people.data
    y = lfw_people.target
    target_names = lfw_people.target_names
    n_classes = target_names.shape[0]

    # 2. split data into training and test data
    x_train, x_test, y_train, y_test = train_test_split(
        x, y, test_size=0.20, random_state=42
    )

    # 3. use 'StandardScaler' on train data and scale both train and test data
    scaler = StandardScaler()
    scaler.fit(x_train)
    x_train = scaler.transform(x_train)
    x_test = scaler.transform(x_test)

    # 7. use `explained_var` function to calculate minimum number of components
    # needed to capture 90% of the variance; print this number
    cumulative_explained_var = explained_var(x_train)
    n_components_90_variance = int(np.argmax(cumulative_explained_var >= 0.90)) + 1
    print(
        f"Number of components needed to capture 90% of explained variance: {n_components_90_variance}"
    )

    # 12. import or paste your cv_svm function above,
    # use it together with computed number of components
    # to call 'pca_train' and train model with reduced number of features,
    # measure time duration
    n_components = n_components_90_variance  # 75
    t0 = time()
    pca, model_pca = pca_train(x_train, y_train, n_components, cv_svm)
    print("done in %0.3fs" % (time() - t0))

    # 13. transform input test data using PCA transform
    X_test_pca = pca.transform(x_test)
    # 14. compute and print accuracy of best model on test set
    y_pred = model_pca.predict(X_test_pca)
    print(
        "The PCA model is {}% accurate".format(
            (accuracy_score(y_pred, y_test) * 100).round(1)
        )
    )

    # 15. use function 'cv_svm' to perform hyperparameter search with cross validation
    # on non-preprocessed train set, measure time
    t0 = time()
    model = cv_svm(x_train, y_train)
    print("done in %0.3fs" % (time() - t0))

    # 16. compute and print accuracy of best model on test set
    y_pred = model.predict(x_test)
    print(
        "The model is {}% accurate".format(
            (accuracy_score(y_pred, y_test) * 100).round(1)
        )
    )

    # 17. (optional) plot top 12 eigenfaces
    eigenfaces = pca.components_.reshape((n_components, h, w))
    eigenface_titles = ["eigenface %d" % i for i in range(eigenfaces.shape[0])]
    plot_image_matrix(eigenfaces, eigenface_titles, h, w)
    plt.show()

    # 18. (optional) plot ROC curves
    plot_roc(model_pca, X_test_pca, y_test, n_classes, target_names)
    plot_roc(model, x_test, y_test, n_classes, target_names)

    # (optional)
    # 24. generate list of parameters to perform grid search
    numbers = np.arange(n_components_90_variance - 10, n_components_90_variance + 11, 5)
    # 25. call 'gs_pca' to determine best number of components and print it
    best_num_comp = gs_pca(x, y, numbers)
    print("Best number of components: ", best_num_comp)

    # 26. repeat steps 12 to 14
    n_components = best_num_comp  # 70

    t0 = time()
    pca, model_pca = pca_train(x_train, y_train, n_components, cv_svm)
    print("done in %0.3fs" % (time() - t0))

    X_test_pca = pca.transform(x_test)
    # compute and print accuracy of best model on test set
    y_pred = model_pca.predict(X_test_pca)
    print(
        "The pca model is {}% accurate".format(
            (accuracy_score(y_pred, y_test) * 100).round(1)
        )
    )
    # calculate and plot ROC curve and AUC
    plot_roc(model_pca, X_test_pca, y_test, n_classes, target_names)
